@@ -18,9 +18,11 @@ from base64 import encodebytes
 
 import ipywidgets
 import numpy as np
+from IPython.display import display
 from traitlets import Bool, Dict, Int, Unicode
 
 from ._png import array2png
+from ._utils import RFBOutputContext, Snapshot
 
 
 @ipywidgets.register
@@ -68,16 +70,70 @@ class RemoteFrameBuffer(ipywidgets.DOMWidget):
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
+        self._ipython_display_ = None  # we use _repr_mimebundle_ instread
+        # Setup an output widget, so that any errors in our callbacks
+        # are actually shown. We display the output in the cell-output
+        # corresponding to the cell that instantiates the widget.
+        self._output_context = RFBOutputContext()
+        display(self._output_context)
         # Init attributes for drawing
         self._rfb_draw_requested = False
         self._rfb_frame_index = 0
         self._rfb_last_confirmed_index = 0
+        self._rfb_last_resize_event = None
         # Init stats
         self.reset_stats()
         # Setup events
         self.on_msg(self._rfb_handle_msg)
         self.observe(self._rfb_schedule_maybe_draw, names=["frame_feedback"])
 
+    def _repr_mimebundle_(self, **kwargs):
+        data = {}
+
+        # Always add plain text
+        plaintext = repr(self)
+        if len(plaintext) > 110:
+            plaintext = plaintext[:110] + "…"
+        data["text/plain"] = plaintext
+
+        # Get the actual representation
+        try:
+            data.update(super()._repr_mimebundle_(**kwargs))
+        except Exception:
+            # On 7.6.3 and below, _ipython_display_ is used instead of _repr_mimebundle_.
+            # We fill in the widget representation that has been in use for 5+ years.
+            data["application/vnd.jupyter.widget-view+json"] = {
+                "version_major": 2,
+                "version_minor": 0,
+                "model_id": self._model_id,
+            }
+
+        # Add initial snapshot. It would be awesome if, when the
+        # notebook is offline, this representation is used instead of
+        # application/vnd.jupyter.widget-view+json. And in fact, Gihub's
+        # renderer does this. Unfortunately, nbconvert still selects
+        # the widget mimetype.
+        # So instead, we display() the snapshot right in front of the
+        # actual widget view, and when the widget view is created, it
+        # hides the snapshot. Ha! That way, the snapshot is
+        # automatically shown when the widget is not loaded!
+        if self._view_name is not None:
+            # data["text/html"] = self.snapshot()._get_html()
+            display(self.snapshot(None, _initial=True))
+
+        return data
+
+    def print(self, *args, **kwargs):
+        """Print to the widget's output area (For debugging purposes).
+
+        In Jupyter, print calls that occur in a callback or an asyncio task
+        may (depending on your version of the notebook/lab) not be shown.
+        Inside ``get_frame()`` and ``handle_event()`` you can use this method
+        instead. The signature of this method is fully compatible with
+        the builtin print function (except for the ``file`` argument).
+        """
+        self._output_context.print(*args, **kwargs)
+
     def close(self, *args, **kwargs):
         """Close all views of the widget and emit a close event."""
         # When the widget is closed, we notify by creating a close event. The
@@ -89,11 +145,57 @@ class RemoteFrameBuffer(ipywidgets.DOMWidget):
         """Receive custom messages and filter our events."""
         if "event_type" in content:
             if content["event_type"] == "resize":
+                self._rfb_last_resize_event = content
                 self.request_draw()
-            self.handle_event(content)
+            elif content["event_type"] == "close":
+                self._repr_mimebundle_ = None
+            with self._output_context:
+                self.handle_event(content)
 
     # ---- drawing
 
+    def snapshot(self, pixel_ratio=None, _initial=False):
+        """Create a snapshot of the current state of the widget.
+
+        Returns an ``IPython DisplayObject`` that can simply be used as
+        a cell output. The display object has a ``data`` attribute that holds
+        the image array data (typically a numpy array).
+        """
+        # Start with a resize event to the appropriate pixel ratio
+        ref_resize_event = self._rfb_last_resize_event
+        if ref_resize_event:
+            w = ref_resize_event["width"]
+            h = ref_resize_event["height"]
+        else:
+            pixel_ratio = pixel_ratio or 1
+            css_width, css_height = self.css_width, self.css_height
+            w = float(css_width[:-2]) if css_width.endswith("px") else 500
+            h = float(css_height[:-2]) if css_height.endswith("px") else 300
+        if pixel_ratio:
+            evt = {
+                "event_type": "resize",
+                "width": w,
+                "height": h,
+                "pixel_ratio": pixel_ratio,
+            }
+            self.handle_event(evt)
+        # Render a frame
+        array = self.get_frame()
+        # Reset pixel ratio
+        if ref_resize_event and pixel_ratio:
+            self.handle_event(ref_resize_event)
+        # Create snapshot object
+        if array is None:
+            array = np.ones((1, 1), np.uint8) * 127
+        if _initial:
+            title = "initial snapshot"
+            class_name = "initial-snapshot-" + self._model_id
+        else:
+            title = "snapshot"
+            class_name = "snapshot-" + self._model_id
+
+        return Snapshot(array, w, h, title, class_name)
+
     def request_draw(self):
         """Schedule a new draw when the widget is ready for it.
 
@@ -123,9 +225,10 @@ class RemoteFrameBuffer(ipywidgets.DOMWidget):
         frames_in_flight = self._rfb_frame_index - feedback.get("index", 0)
         if self._rfb_draw_requested and frames_in_flight < self.max_buffered_frames:
             self._rfb_draw_requested = False
-            array = self.get_frame()
-            if array is not None:
-                self._rfb_send_frame(array)
+            with self._output_context:
+                array = self.get_frame()
+                if array is not None:
+                    self._rfb_send_frame(array)
 
     def _rfb_send_frame(self, array):
         """Actually send a frame over to the client."""
